/*******************************************************************************
*
* E M B E D D E D   W I Z A R D   P R O J E C T
*
*                                                Copyright (c) TARA Systems GmbH
*                                    written by Paul Banach and Manfred Schweyer
*
********************************************************************************
*
* This software is delivered "as is" and shows the usage of other software
* components. It is provided as an example software which is intended to be
* modified and extended according to particular requirements.
*
* TARA Systems hereby disclaims all warranties and conditions with regard to the
* software, including all implied warranties and conditions of merchantability
* and non-infringement of any third party IPR or other rights which may result
* from the use or the inability to use the software.
*
********************************************************************************
*
* DESCRIPTION:
*   This file is part of the interface (glue layer) between an Embedded Wizard
*   generated UI application and the graphics subsystem.
*
*   This template is responsible to access the graphics subsystem within an
*   embedded Linux environment. The display is accessed via fbdev.
*
*   Important: This file is intended to be used as a template. Please adapt the
*   implementation and declarations according your particular hardware.
*
*******************************************************************************/

/* undefine __unix__ to avoid that X11.h is included and EGLNative types are
   defined with X11 types */
#if defined(__unix__)
  #undef __unix__
#endif

#include "ewconfig.h"
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#include <linux/input.h>
#include <linux/fb.h>

#include "gfx_system.h"


void EwPrint( const char* aFormat, ... );



static void*                       FrameBufferAddr    = 0;
static void*                       DoubleBufferAddr   = 0;
static int                         DisplayWidth       = 0;
static int                         DisplayHeight      = 0;

static int                         FbDev              = 0;
static struct fb_fix_screeninfo    FixInfo;
static struct fb_var_screeninfo    VarInfo;
static struct fb_var_screeninfo    VarInfoOrig;

static char*                       DefaultFbDevName   = "/dev/fb0";
static char*                       FbDevName          = NULL;


/*******************************************************************************
* FUNCTION:
*   GfxSystemInit
*
* DESCRIPTION:
*   The function GfxSystemInit is responsible to get an access to the
*   graphics subsystem. In case that the underlying graphics subsystem contains
*   a window manager, the given size is used to create a suitable window.
*   Otherwise, the given size is ignored, because the size of the frame buffer
*   cannot be changed.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   Returns 1 if successful, 0 otherwise.
*
*******************************************************************************/
int GfxSystemInit( void )
{
  unsigned long    bufferSize;

  if (( FbDevName = getenv( "EW_FBDEVICE" )) == NULL )
    FbDevName = DefaultFbDevName;

   /* open the framebuffer device */
  FbDev = open( FbDevName, O_RDWR );
  if ( FbDev < 0 )
  {
    EwPrint( "EwBspDisplayInit: failed to open framebuffer device.\n" );
    return 0;
  }

  /* store current screen settings */
  ioctl( FbDev, FBIOGET_VSCREENINFO, &VarInfoOrig );

  /* query variable framebuffer information */
  ioctl( FbDev, FBIOGET_VSCREENINFO, &VarInfo );

  /* check if required color format is supported by linux framebuffer */
  VarInfo.bits_per_pixel = EW_FRAME_BUFFER_DEPTH * 8;
  ioctl( FbDev, FBIOPUT_VSCREENINFO, &VarInfo );

  ioctl( FbDev, FBIOGET_VSCREENINFO, &VarInfo );

  if ( VarInfo.bits_per_pixel / 8 != EW_FRAME_BUFFER_DEPTH )
  {
    EwPrint( "GfxSystemInit: Color format %s not supported by linux framebuffer!",
      EW_FRAME_BUFFER_COLOR_FORMAT_STRING );
    close( FbDev );
    return 0;
  }

  DisplayWidth  = VarInfo.xres;
  DisplayHeight = VarInfo.yres;

  /* query fixed framebuffer information */
  ioctl( FbDev, FBIOGET_FSCREENINFO, &FixInfo );

  /* determine size of framebuffer in bytes */
  bufferSize = DisplayHeight * FixInfo.line_length;

  /* move the virtual window to top/left corner */
  VarInfo.xoffset = 0;
  VarInfo.yoffset = 0;
  ioctl( FbDev, FBIOPAN_DISPLAY, &VarInfo );

  //EwPrint( "Framebuffer xres: %d yres: %d bits_per_pixel: %d line_length: %d\n",
  //  VarInfo.xres, VarInfo.yres, VarInfo.bits_per_pixel, FixInfo.line_length );

  /* determine address of front and backbuffer */
  #if EW_USE_DOUBLE_BUFFER == 1

    /* double the virtual size of the framebuffer */
    VarInfo.xres_virtual = DisplayWidth;
    VarInfo.yres_virtual = DisplayHeight * 2;
    ioctl( FbDev, FBIOPUT_VSCREENINFO, &VarInfo );

    /* get address of frontbuffer (upper half) and backbuffer (lower half) */
    FrameBufferAddr = mmap( 0, bufferSize * 2, PROT_READ | PROT_WRITE, MAP_SHARED, FbDev, 0 );
	  DoubleBufferAddr = FrameBufferAddr + bufferSize;

	#else

    FrameBufferAddr = mmap( 0, bufferSize, PROT_READ | PROT_WRITE, MAP_SHARED, FbDev, 0 );
	  DoubleBufferAddr = 0;

  #endif

  if (( FrameBufferAddr == 0 ) || ( FrameBufferAddr == MAP_FAILED ))
  {
    EwPrint( "GfxSystemInit: Cannot access framebuffer!" );
    close( FbDev );
    return 0;
  }

  return 1;
}


/*******************************************************************************
* FUNCTION:
*   GfxSystemDone
*
* DESCRIPTION:
*   The function GfxSystemDone is responsible to close the access to the
*   graphics subsystem.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void GfxSystemDone( void )
{
  /* clear frame buffer */
  #if EW_USE_DOUBLE_BUFFER == 1
    memset( FrameBufferAddr, 0, DisplayHeight * FixInfo.line_length * 2 );
  #else
    memset( FrameBufferAddr, 0, DisplayHeight * FixInfo.line_length );
  #endif

  /* restore screen settings */
  ioctl( FbDev, FBIOPUT_VSCREENINFO, &VarInfoOrig );

  /* close the framebuffer device */
  close( FbDev );
}


/*******************************************************************************
* FUNCTION:
*   GfxSystemProcess
*
* DESCRIPTION:
*   The function GfxSystemProcess drives the graphics subsystem or window
*   manager in order to process all pending events.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   1, if further processing is needed, 0 otherwise.
*
*******************************************************************************/
int GfxSystemProcess( void )
{
  return 1;
}


/*******************************************************************************
* FUNCTION:
*   GfxSystemGetInfo
*
* DESCRIPTION:
*   The function GfxSystemGetInfo returns the graphic system info struct.
*
* ARGUMENTS:
*   aGfxSystemInfo  - Pointer to return the graphic system info struct.
*
* RETURN VALUE:
*   None.
*
*******************************************************************************/
void GfxSystemGetInfo( XGfxSystemInfo* aGfxSystemInfo )
{
  if ( ! aGfxSystemInfo )
    return;

  memset( aGfxSystemInfo, 0, sizeof( *aGfxSystemInfo ));

  aGfxSystemInfo->Type         = GFX_SYSTEM_FBDEV;
  aGfxSystemInfo->Interface    = FbDevName;
  aGfxSystemInfo->FrameBuffer  = (void*)(intptr_t)FrameBufferAddr;

  #if EW_USE_DOUBLE_BUFFER == 1
  aGfxSystemInfo->DoubleBuffer = (void*)(intptr_t)DoubleBufferAddr;
  #endif

  aGfxSystemInfo->Pitch        = FixInfo.line_length;
  aGfxSystemInfo->Width        = DisplayWidth;
  aGfxSystemInfo->Height       = DisplayHeight;
}


/*******************************************************************************
* FUNCTION:
*   GfxSystemUpdateSurface
*
* DESCRIPTION:
*   The function GfxSystemUpdateSurface updates the graphic surface.
*
* ARGUMENTS:
*   aSurfaceContext  - Surface context
*   aX1              - Surface update area left.
*   aY1              - Surface update area topt.
*   aX2              - Surface update area right.
*   aY2              - Surface update area bottom.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void GfxSystemUpdateSurface ( void * aSurfaceContext, int aX1, int aY1, int aX2, int aY2 )
{
  #if EW_USE_DOUBLE_BUFFER == 1

    /* in case of linux framebuffer device, double-buffering is handled by shifting
       the framebuffer between the upper and lower half */
    if ( aSurfaceContext == DoubleBufferAddr )
   		VarInfo.yoffset = DisplayHeight;
    else
  		VarInfo.yoffset = 0;

  	/* pan to the current buffer */
  	ioctl( FbDev, FBIOPAN_DISPLAY, &VarInfo );

  #endif
}


/*******************************************************************************
* FUNCTION:
*   GfxSystemWaitForCompletion
*
* DESCRIPTION:
*   The function GfxSystemWaitForCompletion is called to ensure that all pending
*   activities of the display system are completed, so that the rendering of the
*   next frame can start.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void GfxSystemWaitForCompletion( void )
{
  #if EW_USE_DOUBLE_BUFFER == 1

  	/* wait for v-sync */
  	ioctl( FbDev, FBIO_WAITFORVSYNC, 0 );

  #endif
}


/*******************************************************************************
* FUNCTION:
*   GfxSystemSetClut
*
* DESCRIPTION:
*   The function EwBspDisplaySetClut is used to update the hardware CLUT of the
*   current framebuffer.
*
* ARGUMENTS:
*   aClut - Pointer to a color lookup table with 256 entries.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void GfxSystemSetClut( unsigned long* aClut )
{
  /* copy the given color values into your hardware CLUT */
  /* in case of linux framebuffer device, a CLUT with 16bits per color component
     has to be prepared and loaded into hardware via ioctl */
  struct fb_cmap colors;
  unsigned short reds[256],greens[256],blues[256];
  int i;

  colors.start  = 0;
  colors.len    = 256;
  colors.red    = reds;
  colors.green  = greens;
  colors.blue   = blues;
  colors.transp = NULL;

  for ( i = 0; i < 256; i++ )
  {
    unsigned long c = *aClut++;
    reds[i]   = ((c & 0x00FF0000) >> 16) | ((c & 0x00FF0000) >> 8);
    greens[i] = ((c & 0x0000FF00) >>  8) | ((c & 0x0000FF00) >> 0);
    blues[i]  = ((c & 0x000000FF) >>  0) | ((c & 0x000000FF) << 8);
  }

  /* load the palette with the color values */
  ioctl( FbDev, FBIOPUTCMAP, &colors );
}


/* mli, msy */
